#include "tile/base/option.h"

#include <utility>

#include "tile/base/option/option_service.h"

namespace tile {

namespace option {

void
InitializeOptions()
{
    option::OptionService::Instance()->ResolveAll();
}

void
ShutdownOptions()
{
    option::OptionService::Instance()->Shutdown();
}

void
SynchronizeOptions()
{
    option::OptionService::Instance()->UpdateOptions();
}

Json::Value
DumpOptions()
{
    return option::OptionService::Instance()->Dump();
}

}// namespace option

}// namespace tile
